
#
# Aggregate functions (based on structural recursion) on EIterators in Python
#
# Wong Limsoon
# 9 Nov 2020
#

import sys
from sri import Sri
from synciterators import EIterator


#
# Aggregate functions defined in terms of structural recursion.
#

def keep():

    """ A do-nothing aggregate that just returns its input as a list. """

    def comb(acc, y):
        acc.append(y)
        return acc

    zero = lambda: []
    iter = lambda x: x
    return Sri(zero, iter, comb)



def minimize(f = lambda x: x, g = lambda x: x):

    """ Return the minimum value of f(x) over all x in the input,
        along with [ g(x) for x in the input].
    """
        
    def comb(acc, y):
        if y[0] > acc[0]: return acc
        elif y[0] == acc[0]: acc[1].append(y[1]); return acc
        else: return (y[0], [y[1]])
    zero = lambda: (sys.float_info.max, [])
    h = f if not isinstance(f, str) else lambda r: r[f]
    iter = lambda x: (h(x), g(x))
    return Sri(zero, iter, comb)



def maximize(f = lambda x: x, g = lambda x: x):

    """ Return the maximum value of f(x) over all x in the input,
        along with [ g(x) for x in the input].
    """
        
    def comb(acc, y):
        if y[0] < acc[0]: return acc
        elif y[0] == acc[0]: acc[1].append(y[1]); return acc
        else: return (y[0], [y[1]])
    zero = lambda: (sys.float_info.min, [])
    h = f if not isinstance(f, str) else lambda r: r[f]
    iter = lambda x: (h(x), g(x))
    return Sri(zero, iter, comb)

  

def sum(f = lambda x: x): 

    """ sum of f(x) over all x in the input. """

    iter = f if not isinstance(f, str) else lambda r: r[f]
    return Sri(lambda: 0, iter, lambda acc, y: acc + y)
 

def count(): 

    """ Count the items in the input. """

    return sum(lambda x: 1)


def average(f = lambda x: x): 

    """ Average of f(x) over all x in the input. """

    def comb(acc, x): 
        acc[0] += x
        acc[1] += 1
        return acc
    def done(acc): return acc[0]/acc[1] if acc[1] != 0 else 0
    iter = f if not isinstance(f, str) else lambda r: r[f]
    return Sri(lambda: [0, 0], iter, comb, done) 


def smallest(f = lambda x: x):

    """ Smallest f(x) over all x in the input. """

    zero = sys.float_info.max
    iter = f if not isinstance(f, str) else lambda r: r[f]
    return Sri(lambda: zero, iter, lambda acc, x: acc if acc < x else x)


def biggest(f = lambda x: x):

    """ Biggest f(x) over all x in the input. """

    zero = sys.float_info.min
    iter = f if not isinstance(f, str) else lambda r: r[f]
    return Sri(lambda: zero, iter, lambda acc, x: acc if acc > x else x)


def stats(f = lambda x: x):

    """ Common statistics about f(x) over all x in the input. """
 
    def comb(acc, x):
        acc[0] += 1                 # count
        acc[1] += x                 # sum
        acc[2] += x * x             # sum squares
        if x < acc[3]:  acc[3] = x  # min
        if x > acc[4]:  acc[4] = x  # max
        return acc

    def mix(acc):
        cnt = acc[0]
        sum = acc[1]
        ssq = acc[2]
        min = acc[3]
        max = acc[4]
        ave = sum/cnt if cnt != 0 else 0
        vre = (ssq + (cnt*ave*ave) - (2*sum*ave)) / (cnt - 1) if cnt > 1 else 0 
        return { "average": ave, "variance": vre, "count": cnt,
                 "sum": sum, "sumsq": ssq, "min": min, "max": max }

    zero = lambda: [0, 0, 0, sys.float_info.max, sys.float_info.min]
    iter = f if not isinstance(f, str) else lambda r: r[f]
    return Sri(zero, iter, comb, mix)



def forall(f = lambda x: x): 

    """ Check whether f(x) holds for all x in the input.  """

    iter = f if not isinstance(f, str) else lambda r: r[f]
    return Sri(lambda: True, iter, lambda acc, x: acc and x)


def thereis(f = lambda x: x):

    """ Check whether f(x) holds some x in the input.  """

    iter = f if not isinstance(f, str) else lambda r: r[f]
    return Sri(lambda: False, iter, lambda acc, x: acc or x)



#
# Functions for applying aggregate functions.
# The aggregate functions are those defined in terms of
# structural recursion, like the ones above.
#

def groupBy(vec, group, aggr):

    """ Group input (vec) based on a function (group).
        Apply aggregate (aggr) per group.
        Do not assume the input is sorted.
    """

    groups = {}
    for c in vec:
        k = group(c)
        groups[k] = aggr.comb(
                      groups[k] if k in groups else aggr.zero(),
                      aggr.iter(c))
    return [ (k, aggr.done(groups[k])) for k in groups ]



def partitionBy(vec, group, aggr):

    """ Group input (vec) based on a function (group).
        Apply aggregate (aggr) per group.
        Assume that the input is sorted.
    """

    def aggrOnGroup(itc, aggr):
        while itc.hasNext():
            gd = itc.next()
            ac = aggr.comb(aggr.zero(), gd[1])
            while (itc.hasNext() and itc.peekahead(1)[0][0] == gd[0]):
                gd = itc.next()
                ac = aggr.comb(ac, gd[1])
            yield (gd[0], aggr.done(ac))

    it = vec if isinstance(vec, EIterator) else EIterator(vec)
    itc = EIterator(((group(c), aggr.iter(c)) for c in it), lambda: it.close())
    return EIterator(aggrOnGroup(itc, aggr), lambda: itc.close())



def aggregateBy(vec, aggr):

    """ Apply aggregate function (aggr) to the input (vec). """

    if vec == []: return []
    elif isinstance(vec, EIterator) and not vec.hasNext(): return []
    else: return [ aggr.doit(vec) ]



def flatAggregateBy(vec, aggr):

    """ Apply aggregate function (aggr) to the input (vec). """

    return aggr.doit(vec)



def combine(aggrs):

    """Combine multiple aggregate functions into a single one.  """

    zs = { j: aggrs[j].zero for j in aggrs }
    ts = { j: aggrs[j].iter for j in aggrs }
    cs = { j: aggrs[j].comb for j in aggrs }
    ds = { j: aggrs[j].done for j in aggrs }
    zero = lambda: {j: zs[j]() for j in aggrs }
    iter = lambda x: { j: ts[j](x) for j in aggrs }
    comb = lambda acc, x: { j: cs[j](acc[j], x[j]) for j in aggrs }
    done = lambda acc: {j: ds[j](acc[j]) for j in aggrs }
    return Sri(zero, iter, comb, done)



def withAcc(aggr):

    """ Apply aggregate function (aggr) and also return items processed. """

    def zero(): return [aggr.zero(), []]
    def iter(x): return x
    def comb(acc, x):
        acc[0] = aggr.comb(acc[0], aggr.iter(x))
        acc[1].append(x)
        return acc
    def done(acc): return (acc[0], acc[1])
    return Sri(zero, iter, comb, done)



class OpG:

    """ Aggregate functions """

    keep            = keep
    minimize        = minimize
    maximize        = maximize
    count           = count
    sum             = sum
    average         = average
    smallest        = smallest
    biggest         = biggest
    stats           = stats
    forall          = forall
    thereis         = thereis

    """ Aggregate function combinators """

    combine         = combine
    withAcc         = withAcc

    """ Grouping functions """

    groupBy         = groupBy
    partitionBy     = partitionBy
    aggregateBy     = aggregateBy
    flatAggregateBy = flatAggregateBy




