
#
# Synchrony GMQL in Python
#
# Emulate GMQL queries on BED files.
#
# Wong Limsoon
# 5 April 2021
#


from simplebed import *
from synciterators import *
import aggriterators as OpG


MATERIALISE = False 
#   Set default output behaviour to lazy streaming.
#   BED files are quite big. So lazy streaming is
#   a good default behaviour. But some less-skillful
#   programmers are not used to this... and make
#   more mistakes.
#
# MATERIALISE = True
#   Set default output behavior to always materialize
#   on disk. This is actually not a good choice for
#   manipulating BED files. But it is "safer" for
#   less-skillful programmers to use.

def mm(bedEIterator):

    """ Set default output behaviour according to the MATERIALISE flag.
    """
    
    bedEFile = TransientBedEFile(bedEIterator)
    return bedEFile.serialized() if MATERIALISE else bedEFile



class BFOut:

    """ Output operators to use with joinR. """

    left = lambda x, y: [x]

    right = lambda x, y: [y]

    intersectGen= lambda renaming: lambda x, y: [
                    Bed({ **x.dict(),
                          **{ renaming(i): y.dict()[i] for i in y.dict() },
                          "chromStart": max(x.chromStart, y.chromStart),
                          "chromEnd": min(x.chromEnd, y.chromEnd)})
                   ] if BedPred(x, y).overlap(1) else []

    intersect = intersectGen(lambda p: "r." + p)

    bothGen = lambda renaming: lambda x, y: [
                Bed({ **x.dict(),
                      **{ renaming(i): y.dict()[i] for i in y.dict() } })
              ]
       
    both = bothGen(lambda p: "r." + p)

    catGen = lambda renaming: lambda x, y: [
                Bed({ **x.dict(),
                      **{ renaming(i): y.dict()[i] for i in y.dict() },
                      "chromStart": min(x.chromStart, y.chromStart),
                      "chromEnd": max(x.chromEnd, y.chromEnd) })
              ]

    cat = catGen(lambda p: "r." + p)



class BFOPlus(TransientBedEFile):
    def __init__(self, it):
        tr = EIterator(
                ( x for x in it 
                      if x.strand != "-" 
                        for _ in [x.__setattr__("strand", "+")] ),
                lambda: it.close())
        TransientBedEFile.__init__(self, tr.distinct())       


class BFOMinus(TransientBedEFile):
    def __init__(self, it):
        tr = EIterator(
                ( x for x in it 
                      if x.strand != "+" 
                        for _ in [x.__setattr__("strand", "-")] ),
                lambda: it.close())
        TransientBedEFile.__init__(self, tr.distinct())       



class BFONull(TransientBedEFile):
    def __init__(self, it):
        tr = EIterator(
                ( x for x in it for _ in [x.__setattr__("strand", ".")] ),
                lambda: it.close())
        TransientBedEFile.__init__(self, tr.distinct())       


class BFOWithStrand(OnDiskBedEFile):
    def __init__(self, it, f):
        # f: EIterator[Bed] -> EFile[Bed]
        bf = TransientBedEFile(it).serialized()
        plusStrand  = f(BFOPlus(bf.eiterator()).eiterator())
        minusStrand = f(BFOMinus(bf.eiterator()).eiterator())
        filename = plusStrand.mergedWith(minusStrand).serialized().filename
        OnDiskBedEFile.__init__(self, filename)


class BFOHisto(TransientBedEFile):
    def __init__(self, it, minmax):
        # minmax: int -> bool
        eit = EIterator(it)
        def mySeek(eit):
            while eit.hasNext() and not minmax(eit.peekahead(1)[0].htIndex):
                eit.next()
            return eit.hasNext()
        def tr(eit):
            while mySeek(eit):
                cur = eit.next()
                cnt = cur.htIndex
                def canMerge(x):
                    nonlocal cur
                    nonlocal cnt
                    return BedPred(cur, x).overlap(0) and cnt == x.htIndex
                while eit.hasNext() and canMerge(eit.peekahead(1)[0]):
                    nxt = eit.next()
                    cur.chromEnd = nxt.chromEnd
                if eit.peekahead(1) == []:
                    yield cur
                else:
                    nxt = eit.peekahead(1)[0]
                    if BedPred(cur, nxt).overlap(0):
                        yield Bed({ ** cur.dict(), "chromEnd": nxt.chromStart})
                    else:
                        yield cur
        TransientBedEFile.__init__(
            self,
            EIterator(tr(eit) if eit.hasNext() else eit, lambda: eit.close())
          )



class BFOSummit(TransientBedEFile):
    def __init__(self, it, minmax):
        # minmax: int -> bool
        eit = BFOHisto(it, minmax).eiterator()
        up  = True
        def isSummit(cur, hasnxt):
            nonlocal up
            if hasnxt == []: tmp = up; up = True; return tmp
            else:
                nxt, = hasnxt 
                if BedPred(cur, nxt).overlap(0):
                    if cur.htIndex > nxt.htIndex:
                        tmp = up; up = False; return tmp
                    else:
                        up = True; return False
                else:
                    tmp = up; up = True; return tmp
        TransientBedEFile.__init__(
          self,
          EIterator(
            (cur for cur in eit if isSummit(cur, eit.peekahead(1))),
            lambda: eit.close()
          )
        ) 
        


class BFOCover(TransientBedEFile):
    def __init__(self, it, minmax):
        # minmax: int -> bool
        eit = BFOHisto(it, minmax).eiterator()
        def tr(eit):
            while eit.hasNext():
                acc = eit.next()
                while eit.hasNext() and BedPred(acc,eit.peekahead(1)[0]).overlap(0):
                    nxt = eit.next()
                    acc.chromEnd = nxt.chromEnd
                yield acc
        TransientBedEFile.__init__(
           self, 
           EIterator(
             tr(eit),
             lambda: eit.close()
           )
         )



class  BFOFlat(TransientBedEFile):
    def __init__(self, it, minmax):
        # minmax: int -> bool
        eit = EIterator(it).filter(lambda b: minmax(b.htIndex)) 
        def tr(eit):
            while eit.hasNext():
                acc = eit.next()
                while eit.hasNext() and BedPred(acc,eit.peekahead(1)[0]).overlap(0):
                    nxt = eit.next()
                    acc.chromStart = min(acc.chromStart, nxt.chromStart)
                    acc.chromEnd = max(acc.chromEnd, nxt.chromEnd)
                yield acc
        TransientBedEFile.__init__(
           self,
           EIterator(
             tr(eit) if eit.hasNext() else eit,
             lambda: eit.close()
           )
         )



class BFOComplement(TransientBedEFile):
    def __init__(self, it):
        tr = EIterator(
               ( Bed({ "chrom":      cur.chrom, 
                       "chromStart": cur.chromEnd,
                       "chromEnd":   nxt.chromStart, 
                       "name":       "",
                       "score":      0,
                       "strand":     cur.strand })
                    for cur in it
                      for nxt in it.peekahead(1)
                        if BedPred(cur, nxt).endBefore() ),
                lambda: it.close()) 
        TransientBedEFile.__init__(self, tr)       
        


class BFOrder:
    plus       = lambda it: BFOPlus(it)
    minus      = lambda it: BFOMinus(it)
    null       = lambda it: BFONull(it)
    withStrand = lambda it, f: BFOWithStrand(it, f)
    histo      = lambda minmax: lambda it: BFOHisto(it, minmax)
    summit     = lambda minmax: lambda it: BFOSummit(it, minmax)
    cover      = lambda minmax: lambda it: BFOCover(it, minmax)
    flat       = lambda minmax: lambda it: BFOFlat(it,minmax)
    complement = lambda it: BFOComplement(it)



class BFCover:

    """ For use with coverR, histogramR, flatR, summitR """

    overlap = lambda y, x: BedPred(y, x).overlap(1)

    overlapStart = lambda y, x: BedPred(y, x.start()).overlap(0)

    coverGenMin = lambda r, hs: [] if hs == [] else [
                     Bed({ "chrom": r.chrom,
                           "chromStart": r.chromStart,
                           "chromEnd": min([h.chromEnd for h in hs]),
                           "name": "",
                           "score": 0,
                           "strand": r.strand,
                           "htIndex": len(hs) })
                  ]

    coverGenMax = lambda r, hs: [] if hs == [] else [
                     Bed({ "chrom": r.chrom,
                           "chromStart": min([h.chromStart for h in hs]),
                           "chromEnd": max([h.chromEnd for h in hs]),
                           "name": "",
                           "score": 0,
                           "strand": r.strand,
                           "htIndex": len(hs) })
                   ]

    atleast = lambda min: lambda n: n >= min

    atmost = lambda max: lambda n: n <= max

    between = lambda min, max: lambda n: min <= n <= max

    exactly = lambda m: lambda n: n == m





def selectR(check = lambda r: True):

    """ Select BED entries in bedEFile satisfying the check function
    #
    # check: Bed -> bool.
    """

    return lambda bedEFile: mm(bedEFile.eiterator().filter(check))



def projectR(proj = {}):

    """ Project selected meta attributes of BED entries in bedEFile  
    #
    # proj: {l1: p1, ..., ln: pn}
    # p1, ..., pn are functions on a BED entry; 
    #   if pi is as a string, it is taken lambda r: r[pi].
    """

    def cvt(f): return f if not isinstance(f, str) else lambda r: r[f]
    def synchroProject(bedEFile):
        it = bedEFile.eiterator()
        tr = EIterator(
               ( Bed({ "chrom": r.chrom, "chromStart": r.chromStart,
                       "chromEnd": r.chromEnd, "name": r.name,
                       "score": r.score, "strand": r.strand,
                       **{ i: cvt(proj[i])(r) for i in proj }}) 
                   for r in it),
               lambda: it.close())
        return mm(tr)
    return synchroProject



def extendR(proj = {}):

    """ Extend BED entries in bedEFile with additional meta attributes
    #
    # proj: {l1: p1, ..., ln: pn}
    # p1, ..., pn are functions on a bed entry;
    #   if pi is as a string, it is taken lambda r: r[pi].
    """

    def cvt(f): return f if not isinstance(f, str) else lambda r: r[f]
    def synchroExtend(bedEFile):
        nonlocal proj
        it = bedEFile.eiterator()
        tr = EIterator(
               ( Bed({**r.dict(), **{ i: cvt(proj[i])(r) for i in proj }})
                   for r in it),
               lambda: it.close())
        return mm(tr)
    return synchroExtend



def partitionByLociR(aggrs = {}):

    """ Group BED entries by loci, and apply aggregate functions per group.
    #
    # aggrs: {l1: aggr1, ..., ln: aggrn}
    # aggr1, ..., aggrn are aggregate function objects,
    #   e.g., OpG.smallest(), OpG.average(), ...
    """

    aggr = OpG.combine(aggrs)
    def synchroPartitionByLoci(bedEFile):
        tp   = OpG.partitionBy(
                 bedEFile.eiterator(),
                 lambda r: (r.chrom, r.chromStart, r.chromEnd),
                 aggr)
        tr   = EIterator(
                 ( Bed({ "chrom": l[0], "chromStart": l[1], "chromEnd": l[2],
                         "name": "", "score": 0, "strand": "+",
                         **r })
                     for l, r in tp ),
                 lambda: tp.close())
        return mm(tr)
    return synchroPartitionByLoci



def partitionByR(group, aggrs = {}):

    """ Group BED entries, and apply aggregate functions per group.
    #
    # group: string 
    #   is a meta attribute of a BED entry; use it in addition to loci to group
    #   BED entries. Then apply the aggregate functions in aggrs to each group.
    # aggrs: {l1: aggr1, ..., ln: aggrn}
    # aggr1, ..., aggrn are aggregate function objects,
    #   e.g., OpG.smallest(), OpG.average(), ...
    """

    aggr = OpG.combine(aggrs)
    def synchroPartitionBy(bedEFile):
        tp   = OpG.partitionBy(
                 bedEFile.eiterator(),
                 lambda r: (r.chrom, r.chromStart, r.chromEnd),
                 OpG.keep())
        tr   = EIterator(
                 ( Bed({ "chrom": l[0], "chromStart": l[1], "chromEnd": l[2],
                         "name": "", "score": 0, "strand": "+", "group": k,
                         **a })
                     for l, v in tp
                       for k, a in OpG.groupBy(v, lambda r: r[group], aggr)),
                 lambda: tp.close())
        return mm(tr)
    return synchroPartitionBy



def mapR(aggrs = {}, ctr = "count"):

    """ Map BED entries in exBedFile to lmBedEFile, apply aggregate functions
        to entries mapped to the same loci.
    #
    # aggrs: { l1:aggr1, ..., ln:aggrn }
    #   where aggri are aggregate function objects.
    # ctr: string
    #   a default aggregate function is to count the number of BED entries
    #   mapped to a loci; ctr is the name of the attribute for storing 
    #   this count.
    """

    aggrs[ctr] = OpG.count()
    aggr = OpG.combine(aggrs)
    def isBefore(y, x): return BedPred(y, x).isBefore()
    def canSee(y, x): return BedPred(y, x).overlap(1)
    def screen(y, x): return BedPred(y, x).sameStrand()
    def iter(r, hs): 
        nonlocal aggr
        return Bed({ **r.dict(), **OpG.flatAggregateBy(hs, aggr) })
    def synchroMap(lmBedEFile, exBedEFile):
        eit = SynchroEIterator(isBefore, canSee, lmBedEFile, exBedEFile, screen)
        return mm(eit.map(iter))
    return synchroMap
 


def differenceR(exact = False):

    """ Return BED entries in lmBedEFile that nothing in exBedEFile maps to.
    #
    # exact: bool
    #   if True, BED entry mapping is based on exact loci matching.
    #   if False, BED entry mapping is based on loci overlap.
    """

    def isBefore(y, x): return BedPred(y, x).isBefore()
    def canSee(y, x): return BedPred(y, x).overlap(1)
    def screen(y, x): 
        bp = BedPred(y, x)
        return bp.sameStrand() and (bp.sameLoci() if exact else True)
    def grpscreen(x, ys): return ys == []
    def iter(r, hs): return r
    def synchroDiff(lmBedEFile, exBedEFiles):
        if isinstance(exBedEFiles, BedEFile): exBedEFiles = [exBedEFiles]
        if len(exBedEFiles) == 0: return lmBedEFile
        exDiff = exBedEFiles[0]
        exRest = exBedEFiles[1:]
        exDiffs = exDiff.mergedWith(*exRest) if exRest else exDiff
        eit = SynchroEIterator(
                isBefore, canSee, lmBedEFile, exDiffs, 
                screen, grpscreen)
        return mm(eit.map(iter))
    return synchroDiff



# Default limit on canSee for joinR.
LIMIT = 100000


def splitGenoPred(ps, ycs, ncs):
    if ps == []: 
        return (ycs, ncs)
    else:
        hd, *tl = ps
        if isinstance(hd,DL) or isinstance(hd,DLE) or isinstance(hd,Overlap): 
            ycs.append(hd)
        else:
            ncs.append(hd)
        return splitGenoPred(tl, ycs, ncs)


def joinR(geno, output = lambda x, y: [x], check  = lambda x, y: True): 

    """ Join BED entries x in lmBedEFile and y in exBedEFile 
    If Genometric predicates in geno all hold on (y, x)
    and check holds on (x, y), then output(x, y).
    """

    global LIMIT
    if not isinstance(geno, list): geno = [geno]
    ycs, ncs = splitGenoPred(geno, [], [])
    if len(ycs) == 0: ycs = [DLE(LIMIT)]
    def screen(y, x):
        nonlocal ncs
        bp = BedPred(y, x)
        return check(x, y) and all(nc.check(bp) for nc in ncs)
    def grpscreen(x, hs): return not hs == []
    def iter(r, hs): return (o for h in hs for o in output(r, h))
    def isBefore(y, x): return BedPred(y, x).isBefore()
    def canSee(y, x): 
        bp = BedPred(y, x)
        return all(yc.check(bp) for yc in ycs)
    def synchroJoin(lmBedEFile, exBedEFile):
        eit = SynchroEIterator(
                isBefore, canSee, lmBedEFile, exBedEFile, 
                screen, grpscreen)
        return mm(eit.ext(iter))
    return synchroJoin



def mapCoverR(
  coverOp, 
  aggrs    = {},
  overlap  = BFCover.overlapStart,
  check    = BFCover.coverGenMin,
  stranded = True):
    def isBefore(y, x): return BedPred(y, x).isBefore()
    def canSee(y, x): return overlap(y, x)
    def screen(y, x): 
        bp = BedPred(y, x)
        return bp.sameStrand and bp.endAfter()
    def synchroMapCover(bedEFiles):
        if not isinstance(bedEFiles, list): bedEFiles = [bedEFiles]
        if not bedEFiles: return BedEFile([])
        first = bedEFiles[0]
        rest  = bedEFiles[1:]
        merged = (first.mergedWith(*rest) if rest else first).serialized()
        def nn(it): return TransientBedEFile(it) if stranded else BFONull(it)
        starts = nn(merged.eiterator().map(lambda r: r.start()))
        ends = nn(merged.eiterator().map(lambda r: r.end()))
        rf = starts.mergedWith(ends.sorted()).distinct().serialized()
        starts.destruct()
        ends.destruct()
        exBedEFile = merged
        lmBedEFile = rf
        tr = SynchroEIterator(
               isBefore, canSee, 
               lmBedEFile, exBedEFile, 
               screen
             ).ext(check)
        coverRes = BFOWithStrand(tr, coverOp) if stranded else coverOp(tr)
        coverRes = coverRes.serialized()
        tr.close()
        rf.destruct()
        if len(aggrs) == 0: 
            aggrRes = coverRes
        else: 
            aggrRes = (mapR(aggrs, "accIndex")(coverRes, merged)).serialized()
            coverRes.destruct()
        merged.destruct()
        return aggrRes
    return synchroMapCover
            


def coverR(minmax, aggrs = {}, stranded = True):
    # minmax: int -> bool
    return mapCoverR(
             BFOrder.cover(minmax), 
             aggrs, 
             BFCover.overlapStart, 
             BFCover.coverGenMin, 
             stranded)


def summitR(minmax, aggrs = {}, stranded = True):
    # minmax: int -> bool
    return mapCoverR(
             BFOrder.summit(minmax), 
             aggrs, 
             BFCover.overlapStart, 
             BFCover.coverGenMin, 
             stranded)


def histoR(minmax, aggrs = {}, stranded = True):
    # minmax: int -> bool
    return mapCoverR(
             BFOrder.histo(minmax), 
             aggrs, 
             BFCover.overlapStart, 
             BFCover.coverGenMin, 
             stranded)


def flatR(minmax, aggrs = {}, stranded = True):
    # minmax: int -> bool
    return mapCoverR(
             BFOrder.flat(minmax), 
             aggrs, 
             BFCover.overlapStart, 
             BFCover.coverGenMax, 
             stranded)


def complementR(stranded = True):
    # minmax: int -> bool
    return mapCoverR(
             BFOrder.complement, 
             {}, 
             BFCover.overlapStart, 
             BFCover.coverGenMax, 
             stranded)



class BFOps:
    selectR          = selectR
    projectR         = projectR
    extendR          = extendR
    partitionByLociR = partitionByLociR
    partitionByR     = partitionByR
    mapR             = mapR
    differenceR      = differenceR
    joinR            = joinR
    mapCoverR        = mapCoverR
    coverR           = coverR
    histoR           = histoR
    summitR          = summitR
    flatR            = flatR
    complementR      = complementR    



#
# "Object-oriented" version of BedEFile to endow it
# with methods selectR, projectR, ...
#

class BedFile(BedEFile):

    def __init__(self, vec):
        if isinstance(vec, TransientBedEFile):
            self.__class__ = TransientBedFile
            TransientBedFile.__init__(self, vec.vec)
        if isinstance(vec, OnDiskBedEFile):
            self.__class__ = OnDiskBedFile
            OnDiskBedFile.__init__(self, vec.filename, vec.protect)
        elif isinstance(vec, BedEFile):
            BedEFile.__init__(self, vec.vec)
        else:
            BedEFile.__init__(self, vec)

    def __onDisk__(self, filename, protect = False):
        return OnDiskBedFile(filename, protect)

    def __transient__(self, vec):
        return TransientBedFile(vec)

    def __mm__(self, bedEFile):
        if isinstance(bedEFile, TransientBedEFile):
            return self.__transient__(bedEFile.vec)
        elif isinstance(bedFile, OnDiskBedEFile):
            return self.__onDisk__(bedEFile.filename, bedEFile.protect)
        else:
            return self.__init__(bedEFile.vec) 

    def selectR(self, check = lambda r: True):
        return self.__mm__(selectR(check)(self))

    def projectR(self, proj = {}):
        return self.__mm__(projectR(proj)(self))

    def extendR(self, proj = {}):
        return self.__mm__(extendR(proj)(self))

    def partitionByLociR(aggrs = {}):
        return self.__mm__(partitionByLociR(aggrs)(self))

    def partitionByR(group, aggrs = {}):
        return self.__mm__(partitionByR(group, aggrs)(self))

    def mapR(others, aggrs = {}, ctr = "count"):
        return self.__mm__(mapR(aggrs, ctr)(self, others))

    def differenceR(others, exact = False):
        return self.__mm__(differenceR(exact)(self, others))

    def joinR(others,
              geno, 
              output = lambda x, y: [x], 
              check = lambda x, y: True):
        return self.__mm__(joinR(geno, output, check)(self, others))

    def coverR(minmax, aggrs = {}, stranded = True):
        return self.__mm__(coverR(minmax, aggrs, stranded)(self))

    def summitR(minmax, aggrs = {}, stranded = True):
        return self.__mm__(summitR(minmax, aggrs, stranded)(self))

    def histoR(minmax, aggrs = {}, stranded = True):
        return self.__mm__(histoR(minmax, aggrs, stranded)(self))

    def flatR(minmax, aggrs = {}, stranded = True):
        return self.__mm__(flatR(minmax, aggrs, stranded)(self))

    def complementR(stranded = True):
        return self.__mm__(complementR(stranded)(self))



class OnDiskBedFile(BedFile, OnDiskBedEFile):
    def __init__(self, filename, protect = True):
      OnDiskBedEFile.__init__(self, filename, protect)


class TransientBedFile(BedFile, TransientBedEFile):
    def __init__(self, vec):
        TransientBedEFile.__init__(self, vec)




