

#
# EFile in Python
#
# Wong Limsoon
# 3 November 2020
#

"""
EFile represents an "enhanced file." An EFile can be a file on disk,
a list in memory, or a "transient" (e)iterator. The eiterator, which
stands for "enhanced iterator," is transient in the sense that it is
use-once-only, like all iterators.  EFile provides functionality
to serialize/deserialize the object that it represents. It also 
provides eiterators on the object that it represents.  

If you serialize a transient EFile, then you can use/re-use the
serialized EFile many times.
"""

from synciterators import EIterator
import heapq 
import itertools



class EFile:

    """ EFile(serializer, deserializer) represents an 'enhanced file.'
 
        serializer is a function which takes a file handle and an EIterator;
        and writes all objects in the EIterator to the file handle.
        serializer is not expected to close the file handle, as a user
        may want to append more items to it later.

        deserialiser is a function which takes a file handle; and reads all 
        objects in the file handle into an EIterator. The deserializer should
        close the file handle after it has read all the objects in it.
    """

    def __init__(self, vec, serializer, deserializer, suffix = "", protect = False):
        self.vec = list(vec)
        self.deserializer = deserializer
        self.serializer = serializer
        self.suffix = suffix
        self.protect = protect

    def __iter__(self):

        """ Make EFile iterable. """

        return self.eiterator()

    def __getitem__(self, n):

        """ Make EFile "sliceable" like a list/array. """

        it = self.eiterator()
        slice = it._EIterator__myIter[n]
        it.close()
        return slice

    def __len__(self):

        """ Endow a len() function on EFile """

        it = self.eiterator()
        n = 0
        while it.hasNext(): 
            it.next()
            n += 1
        it.close()
        return n

    def __onDisk__(self, filename, protect = False):

        """ Inheriting class should override this,
            so that self.__onDisk__() always invoke
            the right OnDiskEFile().
        """ 

        return OnDiskEFile(
                 filename, 
                 self.serializer, self.deserializer, self.suffix,
                 protect)

    def __transient__(self, vec):

        """ Inheriting class should override this,
            so that self.__transient__() always invoke
            the right TransientEFile().
        """ 

        return TransientEFile(
                 vec,
                 self.serializer, self.deserializer, self.suffix)       

    def __preprocessing__(self, e, newname):
        
        """ Function to preprocess element e before saving. """

        return e

    def savedAs(self, newname):

        """ Save this EFile under a new name, after processing elements. """

        processed = ( self.__preprocessing__(e, newname) for e in self )
        return self.__transient__(processed).serialized(newname)

    def eiterator(self):

        """ Return an enhanced iterator on this EFile. """

        return EIterator(self.vec, lambda: ())

    def serialized(self, filename = ""):

        """ Serialize this EFile to disk. """

        from tempfile import mktemp
        unprotect = filename == ""
        if unprotect: filename = mktemp(self.suffix)
        if not filename.endswith(self.suffix): filename += self.suffix
        myFile = open(filename, "w")
        self.serializer(myFile, self.eiterator())
        myFile.close()
        return self.__onDisk__(filename, not unprotect)

    def destruct(self):

        """ Delete EFile, if it is not protected """

        import os
        if not isinstance(self, OnDiskEFile): 
            if isinstance(self.vec, EIterator): self.vec.close()
            return
        if self.protect: return
        if not os.path.exists(self.filename): return
        os.remove(self.filename)

    def deserialized(self):

        """ Parse a file on disk, incrementally record by record. 
            Return an iterator on the parsing results. This is a 
            placeholder. Inheriting class, please override.
        """

        return self.vec

    def slurped(self):

        """ Slurp this EFile into a string. """
 
        import io
        myFile = io.StringIO()
        self.serializer(myFile, self.eiterator())
        tmp = myFile.getvalue()
        myFile.close()
        return tmp

    def filter(self, check):

        """ Use the check function to filter this EFile. """

        return self.__transient__(( x for x in self if check(x) ))       

    def mergedWith(self, *files):

        """ Merge this EFile with other EFiles. All files are assumed sorted """

        tmpfiles = [self, *files]
        eiters   = [ f.eiterator() for f in tmpfiles ]
        merged   = EIterator(
                     heapq.merge(*eiters), 
                     lambda: [ it.close() for it in eiters ]  )
        return self.__transient__(merged)

    def sorted(self, chunkSize = 100000):

        """ Sort this EFile """

        def grouper(it, chunkSize):
            while it.hasNext():
                yield itertools.chain(
                        (it.next(),), 
                        itertools.islice(it, chunkSize - 1)
                      )
        chunks = [ self.__transient__(sorted(chunk)).serialized() 
                     for chunk in grouper(self.eiterator(), chunkSize) ]
        merged = chunks[0].mergedWith(*chunks[1:]).serialized() 
        [ chunk.destruct() for chunk in chunks ]
        return merged
        
    def distinct(self):

        """ Eliminate consecutive duplicates """

        it = self.eiterator()
        tr = EIterator(
               ( e for e in it if [e] != it.peekahead(1)),
               lambda: it.close())
        return self.__transient__(tr)
        

        
class OnDiskEFile(EFile):

    """ OnDiskEFile is EFile constructor for a file on disk. """

    def __init__(self, filename, serializer, deserializer, suffix = "", protect = True ): 
        from os.path import isfile
        EFile.__init__(self, [], serializer, deserializer, suffix, protect)
        if filename.endswith(suffix): self.filename = filename
        elif isfile(filename): self.filename = filename
        else: self.filename = filename + suffix

    def eiterator(self):
        myFile = open(self.filename, "r")
        return self.deserializer(myFile)
        # self.deserializer is expected to close myFile.

    def serialized(self, filename = ""):
        if filename == "": return self
        elif filename == self.filename: return self
        elif filename + self.suffix == self.filename: return self
        else: return EFile.serialized(self, filename)

    def deserialized(self):
        myFile = open(self.filename, "r")
        tmp = list(self.deserializer(myFile))
        myFile.close()
        # self.deserializer is expected to close myFile.
        # But we defensively do a myFile.close() any way.
        return tmp

    def slurped(self):
        myFile = open(self.filename, "r")
        tmp = myFile.read()
        myFile.close()
        return tmp



class TransientEFile(EFile):

    """ TransientEFile is EFile constructor for a transient iterator. """

    def __init__(self, vec, serializer, deserializer, suffix = ""):
        EFile.__init__(self, [], serializer, deserializer, suffix)
        self.vec = vec
        
    def eiterator(self):
        if isinstance(self.vec, EIterator): tmp = self.vec
        else: tmp = EIterator(self.vec, lambda: ())
        return tmp

    def deserialized(self):
        return list(self.vec)

    

#
# Default EFile based on JSON
#
# This is just a simple demo to show how to implement EFile.
#

class JSONCodec:

    import json

    def serializer(fh, it):
        for e in it:
            print(json.dumps(e), file = fh)

    def deserializer(fh):
        return EIterator((json.loads(x) for x in fh), lambda: fh.close())


class TransientJSON(TransientEFile):
    def __init__(self, vec):
        TransientEFile.__init__(
          self, 
          vec, 
          JSONCodec.serializer, 
          JSONCodec.deserializer, 
          ".json"
        )





