
#
# Synchrony GMQL in Python
#
# Emulate GMQL queries on Sample files.
#
# Wong Limsoon
# 19 Feb 2021
#



from samples import *
from bedfileops import BFOps, BFOut, BFCover 
import aggriterators as OpG




# MATERIALISE = False 
#   Set default output behaviour to lazy streaming.
#   BED files are quite big. So lazy streaming is
#   a good default behaviour. But some less-skillful
#   programmers are not used to this... and make
#   more mistakes.
#
MATERIALISE = True
#   Set default output behavior to always materialize
#   on disk. This is actually not a good choice for
#   manipulating BED files. But it is "safer" for
#   less-skillful programmers to use.

def mm(sampleEIterator):

    """ Set default output behaviour according to the MATERIALISE flag.
    #
    # sampleEIterator: SampleEIterator
    # mm(sampleEIterator): SampleEFile
    #
    """
    
    sampleEFile = TransientSampleEFile(sampleEIterator)
    return sampleEFile.serialized() if MATERIALISE else sampleEFile



class SFOut:

    """ Output operator to be used with joinS. """

    left      = lambda s, t: s

    right     = lambda s, t: t

    overwrite = lambda s, t: s.metaExtended({ "r." + i: t[i] for i in t.meta() })




def onRegion(f):

    """ onRegion() applies f() to each BED file of each Sample
    #
    # f: BedEFile -> BedEFile
    # onRegion(f): SampleEFile -> SampleEFile
    #
    """

    def auxRegion(samples):
        it = samples.eiterator().map(lambda s: s.bedFileUpdated(f(s.bedFile)))
        return mm(it)

    return auxRegion
               

def onRegionS(f):

    """ onRegionS() applies f() to each Sample to produce a new BED file for it.
    #
    # f: Sample -> BedEFile -> BedFile
    # onRegionS(f): SampleEFile -> SampleEFile
    #
    """

    def auxRegion(samples):
        it = samples.eiterator().map(lambda s: s.bedFileUpdated(f(s)(s.bedFile)))
        return mm(it)

    return auxRegion
               


def selectS(onSample = lambda sample: True):

    """ selectS() returns samples satisfying onSample()
    #
    # onSample: Sample -> bool
    # selectS(onSample): SampleEFile -> SampleEFile
    #
    """

    def auxSelect(samples):
        it = (s for s in samples if onSample(s))
        return mm(EIterator(it, lambda: it.close()))

    return auxSelect

     

def projectS(onSample = {}):

    """ projectS() applies projection functions in onSample{} to all samples
    #
    # onSample: {l1:p1, ..., ln:pn}
    #   pi: Sample -> Ai are projection functions on sample.
    #   If pi is a string, it is taken as meta attribute name.
    # projectS(onSample): SampleEFile -> SampleEFile
    #
    """

    def cvt(f, s): return s[f] if isinstance(f, str) else f(s)

    def projs(s): return { l: cvt(onSample[l], s) for l in onSample }

    def auxProject(samples):
        it = ( s.metaUpdated(projs(s)) for s in samples )
        return mm(EIterator(it, lambda: it.close()))

    return auxProject


def extendS(onSample = {}):

    """ extendS() extends samples using projection functions in onSample{}
    #
    # onSample: {l1:p1, ..., ln:pn}
    #   pi: Sample -> Ai are projection functions on sample.
    #   If pi is a string, it is taken as meta attribute name.
    # extendS(onSample): SampleEFile -> SampleEFile
    #
    """

    def cvt(f, s): return s[f] if isinstance(f, str) else f(s)

    def projs(s): return { l: cvt(onSample[l], s) for l in onSample }

    def auxExtend(samples):
        it = ( s.metaExtended(projs(s)) for s in samples )
        return mm(EIterator(it, lambda: it.close()))

    return auxExtend
                     


def groupByS(grp, aggrs = {}): 

    """ Group samples by grp. Apply aggregate functions in aggrs to each group.
    #
    # grp: str or Sample -> A
    # aggrs: {l1: aggr1, ..., ln:aggrn}
    #   aggri are aggregate function objects, e.g. OpG.smallest()
    #
    """

    aggr = OpG.combine(aggrs)

    def l(s): return s[grp] if isinstance(grp, str) else grp(s)

    def auxGroupBy(samples):
        it = ( s.metaExtended(meta)
                   for k, acc in OpG.groupBy(samples, l, OpG.withAcc(aggr))
                     for c, group in [acc]
                       for meta in [{ **c, "group": k }]
                         for s in group )
        return mm(EIterator(it, lambda: it.close()))

    return auxGroupBy



def differenceS(exact = False, joinby = []):

    """ differenceS() returns BED regions that dont overlap other samples'
    #
    # exact: bool
    # joinby: str or (Sample, Sample) -> bool
    # differenceS(exact, joinby): SampleEFile -> SampleEFile
    #
    """

    if isinstance(joinby, str): joinby = [joinby]

    def cvt(f, s, t): return s[f] == t[f] if isinstance(f, str) else f(s, t)

    def jb(s, t): return all(cvt(f, s, t) for f in joinby)

    def diffR(s, ts): 
        return BFOps.differenceR(exact)( 
                     s.bedFile, 
                     [t.bedFile for t in ts])

    def auxDifference(samples, otherSamples):
        it = ( s.bedFileUpdated(diffR(s, ts))
                  for s in samples
                    for ts in [otherSamples.filter(lambda t: jb(s, t))]
                      if ts != [] )
        return mm(EIterator(it, lambda: it.close()))

    return auxDifference



def joinS(onRegion = lambda us,vs: us, onSample = SFOut.overwrite, joinby = []):

    """ joinS() joins the BED regions in samples according to onRegion
    #
    # onRegion: (BedEFile, BedEFile) -> BedEFile
    # onSample: (Sample, Sample) -> Sample
    # joinby: str or (Sample, Sample) -> bool
    # joinS(onREgion, onSample, joinby): SampleEFile -> SampleEFile
    #
    """

    if isinstance(joinby, str): joinby = [joinby]

    def cvt(f, s, t): return s[f] == t[f] if isinstance(f, str) else f(s, t)

    def jb(s, t): return all(cvt(f, s, t) for f in joinby)

    def auxJoin(samples, otherSamples):
        these  = samples.serialized()
        others = otherSamples.serialized()
        it = ( onSample(s, t).bedFileUpdated(onRegion(s.bedFile, t.bedFile))
                for s in these
                  for t in others
                    if jb(s, t) )
        return mm(EIterator(it, lambda: it.close()))

    return auxJoin



def mapS(onRegion = BFOps.mapR(), onSample = SFOut.left, joinby = []):

    """ mapS() is joinS(), 
        where OnRegion defaults to mapping BED regions
        in a sample to other samples.
    #
    # onRegion: (BedEFile, BedEFile) -> BedEFile
    # onSample: (Sample, Sample) -> Sample
    # joinby: str or (Sample, Sample) -> bool
    # mapS(onREgion, onSample, joinby): SampleEFile -> SampleEFile
    #
    """

    return joinS(onRegion, onSample, joinby)



def coverS(onRegion = BFOps.coverR(BFCover.atleast(2))):

    """ coverS() applies the cover operation onRegion to all BED files
    #
    # onRegion: list(BedEFile) -> BedEFile
    # coverS(onREgion): (SampleEFile, *str) -> SampleEFile
    #
    """

    def auxCover(samples, groupby = []):
        if isinstance(groupby, str):
            l = lambda s: s[i]
        elif len(groupby) == 0:
            l = lambda s: ""
        else:
            l = lambda s: str([ s[i] for i in groupby ])
        it = ( Sample(onRegion([ b.bedFile for b in acc ]), {"group":m})
                 for m, acc in OpG.groupBy(samples, l, OpG.keep()) )
        return mm(EIterator(it, lambda: it.close()))

    return auxCover



class SFOps:
    onRegion    = onRegion
    onRegionS   = onRegionS
    selectS     = selectS
    projectS    = projectS
    extendS     = extendS
    groupByS    = groupByS
    mapS        = mapS
    differenceS = differenceS
    joinS       = joinS
    coverS      = coverS




#
# SampleFile is an "object-oriented" wrapper to endow 
# SampleEFile with selectS, projectS, etc. as methods
#

class SampleFile(SampleEFile):

    def __init__(self, vec):
        if isinstance(vec, TransientSampleEFile):
            self.__class__ = TransientSampleFile
            TransientSampleFile.__init__(self, vec.vec)
        elif isinstance(vec, OnDiskSampleEFile):
            self.__class__ = OnDiskSampleFile
            OnDiskSampleFile.__init__(self, vec.filename, vec.protect)
        elif isinstance(vec, SampleEFile):
            SampleEFile.__init__(self, vec.vec)
        else:
            SampleEFile.__init__(self, vec)

    def __onDisk__(self, filename, protect = False):
        return OnDiskSampleFile(filename, protect)

    def __transient__(self, vec):
        return TransientSampleFile(vec)

    def __mm__(self, sampleEFile):
        if isinstance(sampleEFile, TransientSampleEFile): 
            return self.__transient__(sampleEFile.vec)
        elif isinstance(sampleEFile, OnDiskSampleEFile):
            return self.__onDisk__(sampleEFile.filename, sampleEFile.protect)
        else:
            return self.__init__(sampleEFile.vec)

    def onRegion(self, f):
        return self.__mm__(onRegion(f)(self))

    def onRegionS(self, f):
        return self.__mm__(onRegionS(f)(self))

    def selectS(self, onSample = lambda sample: True):
        return self.__mm__(selectS(onSample)(self))

    def projectS(self, onSample = {}):
        return self.__mm__(projectS(onSample)(self))

    def extendS(self, onSample = {}):
        return self.__mm__(extendS(onSample)(self))

    def groupByS(self, grp, aggrs = {}):
        return self.__mm__(groupByS(grp, aggrs)(self))

    def differenceS(self, others, exact = False, joinby = []):
        return self.__mm__(differenceS(exact, joinby)(self, others))

    def joinS(self, 
              others,
              onRegion = lambda us,vs: us, 
              onSample = SFOut.overwrite, 
              joinby = []):
        return self.__mm__(joinS(onRegion, onSample, joinby)(self, others))

    def mapS(self, 
             others,
             onRegion = BFOps.mapR(), 
             onSample = SFOut.left, 
             joinby = []):
        return self.joinS(others, onRegion, onSample, joinby)

    def coverS(self, onRegion = BFOps.coverR(BFCover.atleast(2)), groupby = []):
        return self.__mm__(coverS(onRegion)(self, groupby))



class OnDiskSampleFile(SampleFile, OnDiskSampleEFile):
    def __init__(self, filename, protect = True):
        # Requires filename is a file of Samples
        OnDiskSampleEFile.__init__(self, filename, protect)


class OnDiskAltSampleFile(SampleFile, OnDiskAltSampleEFile):
    def __init__(self, dbPath, dbList):
        # Requires dbPath & dbList point to Stefano-formatted Sample files.
        OnDiskAltSampleEFile.__init__(self, dbPath, dbList)


class TransientSampleFile(SampleFile, TransientSampleEFile):
    def __init__(self, vec):
        # Requires vec is a list/eiterator of Samples
        TransientSampleEFile.__init__(self, vec)


