
#
# Sample and SampleEFile in Python
#
# Wong Limsoon
# 9 Nov 2020
#

from simplebed import Cvt, OnDiskBedEFile
from fileiterators import EFile, OnDiskEFile, TransientEFile
from synciterators import EIterator
import itertools



class Sample:

    """ Sample is a simple representation of a sample.

        A Sample has just two components: 
        - meta, which is a dictionary mapping attributes to their values.
        - bedFile, which is an EFile representing a BED file.
    """

    def __init__(self, bedFile, meta):
        # meta is a dictionary
        # bedFile is an EFile representing a BED file
        for k in meta: setattr(self, k, meta[k])
        self.bedFile = bedFile

    def __getitem__(self, item):
        return getattr(self, item)

    def meta(self):
        d = dict(self.__dict__)
        del d["bedFile"]
        return d

    def bedFileUpdated(self, bedFile):
        return Sample(bedFile, self.meta())
 
    def metaUpdated(self, meta):
        return Sample(self.bedFile, meta)

    def metaExtended(self, meta):
        return Sample(
                 self.bedFile, 
                 self.meta() if meta == {} else { **self.meta(), **meta })

    def trackSavedAs(self, filename):
        bedFile = self.bedFile.serialized(filename)
        return Sample(bedFile, self.meta())

    def trackSorted(self):
        bedFile = self.bedFile.sorted()
        return Sample(bedFile, self.meta())
        
    def trackSavedWithPrefix(self, prefix):
        import pathlib, tempfile
        if isinstance(self.bedFile, OnDiskBedEFile):
            oldpath = pathlib.Path(self.bedFile.filename)
        else:
            oldpath = pathlib.Path(tempfile.mktemp(self.bedFile.suffix))
        newname = prefix + "-" + oldpath.name
        return self.trackSavedAs(newname)
        


class SampleEFileCodec:

    """ Serializer/deserializer for Sample EFile.
    """

    def deserializer(self, fh):

        """ deserializer() is deserializer for Limsoon's sample file.

            Limsoon's sample file is a simple text file, listing sample
            records. The sample records are separated by blank lines.
            Each sample record begins with the filename of the BED file
            associated with the sample, followed on separate lines
            by the sample's meta attributes in the form
            <attribute name>\t<value>.
        """

        def it():
            for key, group in itertools.groupby(fh, lambda x: x.strip() == ""):
                if not key:
                    ls = list(group)
                    sid = ls[0].strip()
                    fs = [(f, Cvt(v.strip()).cvt()) 
                            for l in ls[1:] 
                              for f, v in [l.split("\t")] ]
                    meta = dict(fs + [("sid", sid)])
                    yield Sample(OnDiskBedEFile(sid), meta)
        return EIterator(it(), lambda: fh.close())

    def stfDeserializer(self, dbPath):

        """ stfDeserializer() is deserializer for Stefano's sample file.

            Stefano's sample file has two parts, a database folder (dbPath)
            and a text file listing the sample ids. The deserializer has to
            first read this text file, then uses the sample ids to access
            the samples in the database folder. Each sample is again in two
            files; the first is a BED file with the name sid, the second is
            a text file with the name sid.meta (this second file lists the
            meta data of the sample in the form <attribute name>\t<value>.
        """

        def parser(fh):
            for l in fh:
                sid = l.strip()
                bf = dbPath + "/" + sid
                mp = bf + ".meta"
                mf = open(mp, "r")
                mm = mf.readlines()
                mf.close()
                fs = [(f, Cvt(v).cvt())
                        for l in mm
                          for field in [ l.strip().split("\t") ]
                            if len(field) == 2
                              for f, v in [ field ] ]
                meta = dict(fs + [("sid", sid)])
                yield Sample(OnDiskBedEFile(bf), meta)

        return lambda fh: EIterator(parser(fh), lambda: fh.close())

    def format(self, sample):

        """ The sample's string representation in sample file, Limsoon's format.
        """

        sid = sample.bedFile.serialized().filename
        meta = sample.meta()
        fields = "\n".join([f + "\t" + str(meta[f]) for f in meta])
        return sid + "\n" + fields + "\n"

    def serializer(self, fh, it):

        """ Write all samples in a list/iterator (it) to a file (fh)
        """

        for sample in it: 
            print(self.format(sample), file = fh)


sampleEFileCodec = SampleEFileCodec()



class SampleEFile(EFile):
    def __init__(self, vec):
        # vec: [Sample]
        EFile.__init__(
          self,
          vec,
          sampleEFileCodec.serializer,
          sampleEFileCodec.deserializer,
          ".sfsav"
        )

    def __onDisk__(self, filename, protect = False):
        return OnDiskSampleEFile(filename, protect)

    def __transient__(self, vec):
        return TransientSampleEFile(vec)

    def __preprocessing__(self, e, newname):
        return e.trackSavedWithPrefix(newname)



class OnDiskSampleEFile(SampleEFile, OnDiskEFile):

    """ OnDiskSampleEFile(filename) is EFile representation of
        Limsoon's sample file.  Limsoon's sample file is the
        default format for Synchrony GMQL.
    """

    def __init__(self, filename, protect = True):
        # Requires filename is a file of Samples.
        OnDiskEFile.__init__(
          self, 
          filename,
          sampleEFileCodec.serializer,
          sampleEFileCodec.deserializer,
          ".sfsav",
          protect
        )



class OnDiskAltSampleEFile(SampleEFile, OnDiskEFile):

    """ OnDiskAltSampleEFile(dbPath, dbList) is EFile
        representation of Stefano's sample file,  Stefano's 
        sample file is the default format for GMQL.
    """

    def __init__(self, dbPath, dbList):
        OnDiskEFile.__init__(
          self,
          dbList,
          sampleEFileCodec.serializer,
          sampleEFileCodec.stfDeserializer(dbPath),
          ".sfsav",
          True
        )



class TransientSampleEFile(SampleEFile, TransientEFile):

    """ TransientSampleEFile(vec) represents a list/iterator (vec)
        of samples as EFile. 
    """

    def __init__(self, vec):
        # Requires vec is a list/eiterator of Samples.
        TransientEFile.__init__(
          self,
          vec,
          sampleEFileCodec.serializer,
          sampleEFileCodec.deserializer,
          ".sfsav"
    )     





