
#
# BedEFile in Python
#
# EFile representation of BED file. 
#
# Wong Limsoon
# 2 April 2021
#


import re
from synciterators import EIterator
from fileiterators import EFile, OnDiskEFile, TransientEFile



class Cvt:
    """ Cvt provides functions for converting strings to int and float.
    """

    __intPattern = re.compile("^[+-]?[0-9]+$")
    __floatPattern = re.compile("^[+-]?[0-9]+(.[0-9])?[0-9]*([eE][+-]?[0-9]+)?$")

    def __init__(self, str):
        self.str = str

    def isInt(self):
        tmp = Cvt.__intPattern.fullmatch(self.str)
        return not tmp == None # and self.str == tmp.group()

    def isFloat(self):
        tmp = Cvt.__floatPattern.fullmatch(self.str)
        return not tmp == None # and self.str == tmp.group()

    def __int__(self, x):
        try:
            return int(x)
        except ValueError:
            return float(x)

    def getCvt(self):
        if self.isInt(): return self.__int__
        elif self.isFloat(): return lambda x: float(x)
        else: return lambda x: x

    def cvt(self):
        if self.isInt(): return int(self.str)
        elif self.isFloat(): return float(self.str)
        else: return self.str



class Bed(object):
    """ Bed represents a BED entry.
    
    Bed(d) converts a dictonary d into a Bed object.
    d is assumed to have the right BED entry attributes,
    viz. chrom, chromStart, chromEnd, name, score, and
    strand, and other optional attributes.
    """

    def __init__(self, args):
        for k in args:
            setattr(self, k, args[k])

    def __getitem__(self, item):
         return getattr(self, item)

    def dict(self): return self.__dict__

    def __eq__(self, other):
        return self.dict() == other.dict() if isinstance(other, Bed) else False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __lt__(self, other): 
        if isinstance(other, Bed):
            if self.chrom < other.chrom:             return True
            elif self.chrom > other.chrom:           return False
            elif self.chromStart < other.chromStart: return True
            elif self.chromStart > other.chromStart: return False
            elif self.chromEnd < other.chromEnd:     return True
            elif self.chromEnd > other.chromEnd:     return False
            elif self.strand > other.strand:         return False
            else: return True
        else:
            return self < other

    def start(self):
        return Bed({ **self.dict(), "chromEnd": self.chromStart })

    def end(self):
        return Bed({ **self.dict(), "chromStart": self.chromEnd })

    def metaExtended(self, meta):
        return Bed({ **self.dict(), **meta})



class BedEFileCodec:
    """ BedEFileCodec provides serializer/deserializer for BED files.

    A standard BED file has six required fields, as well as user-defined
    additional fields.  A standard BED file does not provide names of 
    its data fields in a self-contained way. Users got to give names to
    these fields inside their applications. 

    A self-describing BED file is a modified BED file. Basically,
    the first data line is modified to provide names for each data
    field. Our serializer/deserializer is for a self-describing BED file.
    """
        
    def deserializer(self):

        fields = [ ("chrom",      lambda x: x),
                   ("chromStart", lambda x: int(x)),
                   ("chromEnd",   lambda x: int(x)),
                   ("name",       lambda x: x),
                   ("score",      lambda x: int(x)),
                   ("strand",     lambda x: x) ]

        def field(n, i, v):
            if n > 0 or i < 6:
                fl, cvt = fields[i]
            else:
                fv = v.split("=")
                if len(fv) == 1: 
                    fl = "_" + str(i)
                    cvt = Cvt(v).getCvt()
                    fields.append((fl, cvt))
                else:
                    v = fv[1]
                    fl = fv[0]
                    cvt = Cvt(v).getCvt()
                    fields.append((fl, cvt))
            return (fl, cvt(v))

        def parser(fh):
            n = 0
            def process(l):
                nonlocal n
                e = l.strip().split("\t")
                d = {f:v for i in range(0,len(e)) for f,v in [field(n,i,e[i])]}
                n += 1
                return Bed(d)
            return EIterator(
              (process(l) for l in fh if not l.startswith("track")),
              lambda: fh.close()
            )

        return parser

    def serializer(self, fh, it):

        n = 0

        def head(e):
            ch = str(e.chrom)
            cs = str(e.chromStart)
            ce = str(e.chromEnd)
            nm = str(e.name)
            sc = str(e.score)
            st = str(e.strand)
            dict = e.dict()
            rest = [(i + "=" + str(dict[i])) for i in dict][6:] 
            return "\t".join([ch, cs, ce, nm, sc, st] + rest)

        def format(e): 
            dict = e.dict()
            return "\t".join([str(dict[i]) for i in dict])

        for e in it:
            print(format(e) if n > 0 else head(e), file = fh)
            n += 1


bedEFileCodec = BedEFileCodec()


class BedEFile(EFile):
    def __init__(self, vec):
        # vec: [Bed]
        EFile.__init__(
          self, 
          vec, 
          bedEFileCodec.serializer,
          bedEFileCodec.deserializer(),
          ".bed"
        )

    def __onDisk__(self, filename, protect = False):
        return OnDiskBedEFile(filename, protect)

    def __transient__(self, vec):
        return TransientBedEFile(vec)



class OnDiskBedEFile(BedEFile, OnDiskEFile):
    def __init__(self, filename, protect = True):
        OnDiskEFile.__init__(
          self,
          filename,
          bedEFileCodec.serializer,
          bedEFileCodec.deserializer(),
          ".bed",
          protect
        ) 

               

class TransientBedEFile(BedEFile, TransientEFile):
    def __init__(self, vec):
        TransientEFile.__init__(
          self,
          vec,
          bedEFileCodec.serializer,
          bedEFileCodec.deserializer(),
          ".bed"
        )



class BedPred:
    """ BedPred(bedX, bedY) provides some predicates on a pair of BED entries
    """

    def __init__(self, bedX, bedY):
        self.x        = bedX
        self.y        = bedY
        self.sameChrom   = bedX.chrom == bedY.chrom

    def sameStrand(self):
        return (
          self.sameChrom and 
          (self.x.strand != "-" and self.y.strand != "-") or
          (self.x.strand != "+" and self.y.strand != "+")
        )

    def sameLoci(self):
        return ( 
          self.sameChrom and 
          self.x.chromStart == self.y.chromStart and
          self.x.chromEnd == self.y.chromEnd
        )

    def startBefore(self):
        return self.sameChrom and self.x.chromStart < self.y.chromStart

    def startAfter(self):
        return self.sameChrom and self.x.chromStart > self.y.chromEnd

    def endBefore(self):
        return self.sameChrom and self.x.chromEnd < self.y.chromStart

    def endAfter(self):
        return self.sameChrom and self.x.chromEnd > self.y.chromEnd

    def startTogether(self):
        return self.sameChrom and self.x.chromStart == self.y.chromStart

    def endTogether(self):
        return self.sameChrom and self.x.chromEnd == self.y.chromEnd

    def isBefore(self):
        return self.x.chrom < self.y.chrom or self.startBefore()

    # inside is not antimonotonic
    def inside(self):
        return (
          self.sameChrom and
          self.y.chromStart <= self.x.chromStart and
          self.y.chromEnd >= self.x.chromEnd
        )

    # enclose is not antimonotonic
    def enclose(self):
        return (
          self.sameChrom and
          self.x.chromStart <= self.y.chromStart and
          self.x.chromEnd >= self.y.chromEnd
        )

    # outside is not antimonotonic
    def outside(self):
        return self.sameChrom and not self.inside() and not self.enclose()

    # touch is not antimonotonic
    def touch(self):
        return (
          self.sameChrom and
          (self.x.chromEnd == self.y.chromStart or
           self.y.chromEnd == self.x.chromStart)
        )

    def overlap(self, n = 1):
        return (
          self.sameChrom and
          ((self.y.chromStart <= self.x.chromStart <= self.y.chromEnd) or
           (self.x.chromStart <= self.y.chromStart <= self.x.chromEnd)) and
          (min(self.x.chromEnd, self.y.chromEnd) -
           max(self.x.chromStart, self.y.chromStart)) >= n
        )

    def dist(self):
        from sys import maxsize
        if not self.sameChrom: return maxsize
        elif self.endBefore(): return self.y.chromStart - self.x.chromEnd
        elif self.startAfter(): return self.x.chromStart - self.y.chromEnd 
        else: return 0         

    def dl(self, n):
        return self.dist() < n

    # dg is not antimonotonic
    def dg(self, n):
        return self.dist() > n

    def dle(self, n):
        return self.dist() <= n

    # dge is not antimonotonic
    def dge(self, n):
        return self.dist() >= n

    def cond(self, pred):
        return pred(self.x, self.y)

    def basedOnStart(self):
        return BedPred(x.start, y.start)

    def basedOnEnd(self):
        return BedPred(x.end, y.end)

        

class GenometricPred:
    def __init__(self, cond):
        # cond: BedPred -> bool
        self.cond = cond
 
    def check(self, bedPred): return self.cond(bedPred)

    def andalso(self, genoPred):
        # genoPred: GenometricPred
        return Genometric(
          lambda bedPred: self.check(bedPred) and genoPred.check(bedPred)
        ) 

    def orelse(self, genoPred):
        # genoPred: GenometricPred
        return Genometric(
          lambda bedPred: self.check(bedPred) or genoPred.check(bedPred)
        ) 

    def butnot(self, genoPred):
        # genoPred: GenometricPred
        return Genometric(
          lambda bedPred: self.check(bedPred) and not genoPred.check(bedPred)
        ) 


class SameChrom(GenometricPred):
    def __init__(self):
        GenometricPred.__init__(self, lambda bedPred: bedPred.sameChrom)


class SameStrand(GenometricPred):
    def __init__(self):
        GenometricPred.__init__(self, lambda bedPred: bedPred.sameStrand())


class SameLoci(GenometricPred):
    def __init__(self):
        GenometricPred.__init__(self, lambda bedPred: bedPred.sameLoci())


class EndAfter(GenometricPred):
    def __init__(self):
        GenometricPred.__init__(self, lambda bedPred: bedPred.endAfter())


class EndBefore(GenometricPred):
    def __init__(self):
        GenometricPred.__init__(self, lambda bedPred: bedPred.endBefore())


class StartAfter(GenometricPred):
    def __init__(self):
        GenometricPred.__init__(self, lambda bedPred: bedPred.startAfter())


class StartBefore(GenometricPred):
    def __init__(self):
        GenometricPred.__init__(self, lambda bedPred: bedPred.startBefore())


class StartTogether(GenometricPred):
    def __init__(self):
        GenometricPred.__init__(self, lambda bedPred: bedPred.startTogether())


class EndTogether(GenometricPred):
    def __init__(self):
        GenometricPred.__init__(self, lambda bedPred: bedPred.endTogether())


class IsBefore(GenometricPred):
    def __init__(self):
        GenometricPred.__init__(self, lambda bedPred: bedPred.isBefore())


class Inside(GenometricPred):
    def __init__(self):
        GenometricPred.__init__(self, lambda bedPred: bedPred.inside())


class Enclose(GenometricPred):
    def __init__(self):
        GenometricPred.__init__(self, lambda bedPred: bedPred.enclose())


class Outside(GenometricPred):
    def __init__(self):
        GenometricPred.__init__(self, lambda bedPred: bedPred.outside())


class Touch(GenometricPred):
    def __init__(self):
        GenometricPred.__init__(self, lambda bedPred: bedPred.touch())


class Overlap(GenometricPred):
    def __init__(self, n = 1):
        GenometricPred.__init__(self, lambda bedPred: bedPred.overlap(n))


class DL(GenometricPred):
    def __init__(self, n):
        GenometricPred.__init__(self, lambda bedPred: bedPred.dl(n))


class DG(GenometricPred):
    def __init__(self, n):
        GenometricPred.__init__(self, lambda bedPred: bedPred.DG(n))


class DLE(GenometricPred):
    def __init__(self, n):
        GenometricPred.__init__(self, lambda bedPred: bedPred.dle(n))


class DGE(GenometricPred):
    def __init__(self, n):
        GenometricPred.__init__(self, lambda bedPred: bedPred.dge(n))


class GenPred(GenometricPred):
    def __init__(self, pred):
        GenometricPred.__init__(self, lambda bedPred: bedPred.cond(pred))



