
#
# Structural recursion on Synchrony iterators in Python
#
# Wong Limsoon
# 9 Nov 2020
#

""" 
The standard structural recursion sri is equivalent to 
the 'fold' function. I.e., it is defined by the recursion
equations:
 
sri(zero, comb)[]     = zero
sri(zero, comb)(x::y) = sri(comb(zero, x), comb)(y)

In this module, instead of being a function, sri is implemented
as a 'data structure' recording the components (i and e) that 
define a structural recursion. The data strucrure is endowed with
a method doit(vec) for applying the structural recursion defined by
these components on an input list/iteractor (vec). Thus, this
is a so-called 'trampoline' style implementation that make the
 structural recursion run on a heap instead of a stack.

For efficiency, the structural recursion data structure also
includes preprocessing (iter) and postprocessing (done) components.
So it actually implements a structural recursion defined as:

sri(zero, iter, comb, done)[]     = done(zero)
sri(zero, iter, comb, done)(x::y) = sri(comb(zero, iter(x)), iter, comb, done)(y)
"""

 

from synciterators import EIterator


class Sri:

    def __init__(self, zero, iter, comb, done = lambda x: x):

        """ Sri(zero, iter, comb, done).doit: list[A] -> D
        # 
        # zero: () -> C
        # iter: A  -> B
        # comb: (C, B) -> C
        # done: C -> D
        """

        self.zero = zero
        self.iter = iter
        self.comb = comb
        self.done = done

    def doit(self, vec):
        it = EIterator(vec) if isinstance(vec, list) else vec
        acc = self.zero()
        while it.hasNext():
          i = self.iter(it.next())
          acc = self.comb(acc, i)
        return self.done(acc)

    def compose(self, f):

        """ Compose a postprocessing function to this structural recusion
        """

        return Sri(
                 self.zero,
                 self.iter,
                 self.comb,
                 lambda x: f(self.done(x))
               )

    def combine(self, them, mix):

        """ Combine multiple structural recursions into a single one.
            This avoids processing the input list/iterator multiple
            times. This makes it possible to run multiple structural
            recursions on streaming data (i.e. iterators) which is
            read-once-and-gone.
        #
        # self: this structural recursion
        # them: other structural recursions to be combined with self.
        # mix:  a function for combining the final output of these
        #   multiple structural recursions, as if these recursions
        #   are run independently.
        """

        gs = [self, *them]
        zs = [g.zero for g in gs]
        ts = [g.iter for g in gs]
        cs = [g.comb for g in gs]
        rs = [g.done for g in gs]
        l = len(gs)
        newz = lambda: [z() for z in zs]
        newi = lambda x: [t(x) for t in ts]
        newr = lambda ac: mix([rs[j](ac[j]) for j in range(0, l)])

        def newc(ac, ys):
            for j in range(0, l): ac[j] = cs[j](ac[j], ys[j])
            return ac

        return Sri(newz, newi, newc, newr)



